"""
The MIT License (MIT)

Copyright (c) 2021-present tag-epic

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import os
from typing import TYPE_CHECKING, Optional, Tuple, Type, TypeVar

from ..components import TextInput as TextInputComponent
from ..enums import ComponentType, TextInputStyle
from ..guild import Guild
from ..state import ConnectionState
from ..utils import MISSING
from .item import Item

__all__ = ("TextInput",)

if TYPE_CHECKING:
    from ..types.components import TextInputComponent as TextInputComponentPayload
    from ..types.interactions import ComponentInteractionData
    from .view import View


T = TypeVar("T", bound="TextInput")
V = TypeVar("V", bound="View", covariant=True)


class TextInput(Item[V]):
    """Represent a UI text input.

    .. versionadded:: 2.0

    Parameters
    ----------
    label: :class:`str`
        The label of the text input
    style: :class:`TextInputStyle`
        The style of the text input.
        By default, the style is ``TextInputStyle.short``, a one line
        input, but you can use ``TextInputStyle.paragraph``, a multi line
        input.
    custom_id: Optional[:class:`str`]
        The ID of the text input that get received during an interaction.
    row: Optional[:class:`int`]
        The relative row this text input belongs to. A Discord component can only have 5
        rows. By default, items are arranged automatically into those 5 rows. If you'd
        like to control the relative positioning of the row then passing an index is advised.
        For example, row=1 will show up before row=2. Defaults to ``None``, which is automatic
        ordering. The row number must be between 0 and 4 (i.e. zero indexed).
    min_length: Optional[:class:`int`]
        The minimal length of the user's input
    max_length: Optional[:class:`int`]
        The maximal length of the user's input
    required: Optional[:class:`bool`]
        If ``True``, the user cannot send the form without filling this
        field.
    default_value: Optional[:class:`str`]
        The value already in the input when the user open the form.
    placeholder: Optional[:class:`str`]
        The text shown to the user when the text input is empty.
    """

    __item_repr_attributes__: Tuple[str, ...] = (
        "custom_id",
        "label",
        "style",
        "min_length",
        "max_length",
        "required",
        "default_value",
        "value",
        "placeholder",
    )

    def __init__(
        self,
        label: str,
        *,
        style: TextInputStyle = TextInputStyle.short,
        custom_id: str = MISSING,
        row: Optional[int] = None,
        min_length: Optional[int] = 0,
        max_length: Optional[int] = 4000,
        required: Optional[bool] = None,
        default_value: Optional[str] = None,
        placeholder: Optional[str] = None,
    ):
        self._provided_custom_id = custom_id is not MISSING
        custom_id = os.urandom(16).hex() if custom_id is MISSING else custom_id
        self._underlying = TextInputComponent._raw_construct(
            type=ComponentType.text_input,
            custom_id=custom_id,
            label=label,
            style=style,
            min_length=min_length,
            max_length=max_length,
            required=required,
            value=default_value,
            placeholder=placeholder,
        )
        self.row = row
        self._inputed_value: Optional[str] = None

    @property
    def style(self) -> TextInputStyle:
        """:class:`nextcord.TextInputStyle`: The style of the text input."""
        return self._underlying.style

    @style.setter
    def style(self, value: TextInputStyle) -> None:
        self._underlying.style = value

    @property
    def custom_id(self) -> Optional[str]:
        """Optional[:class:`str`]: The ID of the text input that gets received during an interaction."""
        return self._underlying.custom_id

    @custom_id.setter
    def custom_id(self, value: Optional[str]) -> None:
        if value is not None and not isinstance(value, str):
            raise TypeError("custom_id must be None or str")

        self._underlying.custom_id = value  # type: ignore

    @property
    def label(self) -> str:
        """:class:`str`: The label of the text input."""
        return self._underlying.label

    @label.setter
    def label(self, value: str) -> None:
        if value is None:
            raise TypeError("label must cannot be None")
        self._underlying.label = str(value)

    @property
    def min_length(self) -> Optional[int]:
        """:class:`int`: The minimum input length for a text input"""
        return self._underlying.min_length

    @min_length.setter
    def min_length(self, value: int) -> None:
        self._underlying.min_length = value

    @property
    def max_length(self) -> Optional[int]:
        """:class:`int`: The maximum input length for a text input"""
        return self._underlying.max_length

    @max_length.setter
    def max_length(self, value: int) -> None:
        self._underlying.max_length = value

    @property
    def required(self) -> Optional[bool]:
        """:class:`bool`: Whether this component is required to be filled"""
        return self._underlying.required

    @required.setter
    def required(self, value: Optional[bool]) -> None:
        if value is not None and not isinstance(value, bool):
            raise TypeError("required must be None or bool")

        self._underlying.required = value

    @property
    def default_value(self) -> Optional[str]:
        """Optional[:class:`str`]: The value already in the text input when the user open the form."""
        return self._underlying.value

    @default_value.setter
    def default_value(self, value: Optional[str]) -> None:
        if value is not None and not isinstance(value, str):
            raise TypeError("default_value must be None or str")
        self._underlying.value = value

    @property
    def value(self) -> Optional[str]:
        """Optional[:class:`str`]: The value sent by the user.
        This field is updated when an interaction is received.
        ``TextInput.value`` is ``None`` when no interaction where received.
        """
        return self._inputed_value

    @property
    def placeholder(self) -> Optional[str]:
        """Optional[:class:`str`]: The text shown to the user when the text input is empty."""
        return self._underlying.placeholder

    @placeholder.setter
    def placeholder(self, value: Optional[str]) -> None:
        if value is not None and not isinstance(value, str):
            raise TypeError("placeholder must be None or str")

        self._underlying.placeholder = value

    @property
    def width(self) -> int:
        return 5

    @classmethod
    def from_component(cls: Type[T], text_input: TextInputComponent) -> T:
        return cls(
            style=text_input.style,
            custom_id=text_input.custom_id,
            label=text_input.label,
            min_length=text_input.min_length,
            max_length=text_input.max_length,
            required=text_input.required,
            # value=text_input.value,  # FIXME: figure out what this was
            placeholder=text_input.placeholder,
            row=None,
        )

    @property
    def type(self) -> ComponentType:
        return self._underlying.type

    def to_component_dict(self) -> TextInputComponentPayload:
        return self._underlying.to_dict()

    def is_dispatchable(self) -> bool:
        return True

    def refresh_component(self, text_input: TextInputComponent) -> None:
        self._underlying = text_input

    def refresh_state(
        self, data: ComponentInteractionData, state: ConnectionState, guild: Optional[Guild]
    ) -> None:
        self._inputed_value = data.get("value", "")



# -*- coding: utf-8 -*-

"""
The MIT License (MIT)

Copyright (c) 2015-present Rapptz

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

import asyncio
from datetime import datetime, timedelta
from typing import TYPE_CHECKING, Any, Dict, Generic, List, Optional, Tuple, TypeVar, Union

from . import utils
from .channel import ChannelType, PartialMessageable
from .embeds import Embed
from .enums import InteractionResponseType, InteractionType, try_enum
from .errors import ClientException, HTTPException, InteractionResponded, InvalidArgument
from .file import File
from .member import Member
from .message import Attachment, Message
from .mixins import Hashable
from .object import Object
from .permissions import Permissions
from .user import ClientUser, User
from .utils import snowflake_time
from .webhook.async_ import Webhook, WebhookMessage, async_context, handle_message_parameters

__all__ = (
    "Interaction",
    "InteractionMessage",
    "InteractionResponse",
    "PartialInteractionMessage",
)

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .application_command import BaseApplicationCommand, SlashApplicationSubcommand
    from .channel import CategoryChannel, ForumChannel, StageChannel, TextChannel, VoiceChannel
    from .client import Client
    from .guild import Guild
    from .message import AllowedMentions
    from .state import ConnectionState
    from .threads import Thread
    from .types.interactions import Interaction as InteractionPayload, InteractionData
    from .ui.modal import Modal
    from .ui.view import View

    InteractionChannel = Union[
        VoiceChannel,
        StageChannel,
        TextChannel,
        CategoryChannel,
        Thread,
        PartialMessageable,
        ForumChannel,
    ]

MISSING: Any = utils.MISSING

ClientT = TypeVar("ClientT", bound="Client")


class InteractionAttached(dict):
    """Represents the attached data of an interaction.

    This is used to store information about an :class:`Interaction`. This is useful if you want to save some data from a :meth:`ApplicationCommand.application_command_before_invoke` to use later in the callback.

    Example
    -------

    .. code-block:: python3

        async def attach_db(interaction: Interaction):
            interaction.attached.db = some_real_database()

        async def release_db(interaction: Interaction):
            interaction.attached.db.close()

        @bot.slash_command()
        @application_checks.before_invoke(attach_db)
        @application_checks.after_invoke(release_db)
        async def who(interaction: Interaction): # Output: <User> used who at <Time>
            data = interaction.attached.db.get_data(interaction.user.id)
            await interaction.response.send_message(data)
    """

    def __init__(self):
        super().__init__()
        self.__dict__ = self

    def __repr__(self):
        return f"<InteractionAttached {super().__repr__()}>"


class Interaction(Hashable, Generic[ClientT]):
    """Represents a Discord interaction.

    An interaction happens when a user does an action that needs to
    be notified. Current examples are slash commands and components.

    .. container:: operations

        .. describe:: x == y

            Checks if two interactions are equal.

        .. describe:: x != y

            Checks if two interactions are not equal.

        .. describe:: hash(x)

            Returns the interaction's hash.

    .. versionadded:: 2.0

    .. versionchanged:: 2.1
        :class:`Interaction` is now hashable.

    Attributes
    ----------
    id: :class:`int`
        The interaction's ID.
    type: :class:`InteractionType`
        The interaction type.
    guild_id: Optional[:class:`int`]
        The guild ID the interaction was sent from.
    channel_id: Optional[:class:`int`]
        The channel ID the interaction was sent from.
    locale: Optional[:class:`str`]
        The users locale.
    guild_locale: Optional[:class:`str`]
        The guilds preferred locale, if invoked in a guild.
    application_id: :class:`int`
        The application ID that the interaction was for.
    user: Optional[Union[:class:`User`, :class:`Member`]]
        The user or member that sent the interaction.
    message: Optional[:class:`Message`]
        The message that sent this interaction.
    token: :class:`str`
        The token to continue the interaction. These are valid
        for 15 minutes.
    data: :class:`dict`
        The raw interaction data.
    attached: :class:`InteractionAttached`
        The attached data of the interaction. This is used to store any data you may need inside the interaction for convenience. This data will stay on the interaction, even after a :meth:`Interaction.application_command_before_invoke`.
    application_command: Optional[:class:`ApplicationCommand`]
        The application command that handled the interaction.
    """

    __slots__: Tuple[str, ...] = (
        "id",
        "type",
        "guild_id",
        "channel_id",
        "data",
        "application_id",
        "message",
        "user",
        "locale",
        "guild_locale",
        "token",
        "version",
        "application_command",
        "attached",
        "_permissions",
        "_app_permissions",
        "_state",
        "_session",
        "_original_message",
        "_cs_response",
        "_cs_followup",
        "_cs_channel",
    )

    def __init__(self, *, data: InteractionPayload, state: ConnectionState):
        self._state: ConnectionState = state
        self._session: ClientSession = state.http._HTTPClient__session  # type: ignore
        # TODO: this is so janky, accessing a hidden double attribute
        self._original_message: Optional[InteractionMessage] = None
        self.attached = InteractionAttached()
        self.application_command: Optional[
            Union[SlashApplicationSubcommand, BaseApplicationCommand]
        ] = None
        self._from_data(data)

    def _from_data(self, data: InteractionPayload):
        self.id: int = int(data["id"])
        self.type: InteractionType = try_enum(InteractionType, data["type"])
        self.data: Optional[InteractionData] = data.get("data")
        self.token: str = data["token"]
        self.version: int = data["version"]
        self.channel_id: Optional[int] = utils.get_as_snowflake(data, "channel_id")
        self.guild_id: Optional[int] = utils.get_as_snowflake(data, "guild_id")
        self.application_id: int = int(data["application_id"])
        self.locale: Optional[str] = data.get("locale")
        self.guild_locale: Optional[str] = data.get("guild_locale")

        self.message: Optional[Message]
        try:
            message = data["message"]
            self.message = self._state._get_message(int(message["id"])) or Message(
                state=self._state, channel=self.channel, data=message  # type: ignore
            )
        except KeyError:
            self.message = None

        self.user: Optional[Union[User, Member]] = None
        self._app_permissions: int = int(data.get("app_permissions", 0))
        self._permissions: int = 0

        # TODO: there's a potential data loss here
        if self.guild_id:
            guild = self.guild or Object(id=self.guild_id)
            try:
                member = data["member"]
            except KeyError:
                pass
            else:
                cached_member = self.guild and self.guild.get_member(int(member["user"]["id"]))  # type: ignore # user key should be present here
                self.user = cached_member or Member(state=self._state, guild=guild, data=member)  # type: ignore # user key should be present here
                self._permissions = int(member.get("permissions", 0))
        else:
            try:
                user = data["user"]
                self.user = self._state.get_user(int(user["id"])) or User(
                    state=self._state, data=user
                )
            except KeyError:
                pass

    @property
    def client(self) -> ClientT:
        """:class:`Client`: The client that handled the interaction."""
        return self._state._get_client()  # type: ignore

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild the interaction was sent from."""
        return self._state and self._state._get_guild(self.guild_id)

    @property
    def created_at(self) -> datetime:
        """:class:`datetime.datetime`: An aware datetime in UTC representing the creation time of the interaction."""
        return snowflake_time(self.id)

    @property
    def expires_at(self) -> datetime:
        """:class:`datetime.datetime`: An aware datetime in UTC representing the time when the interaction will expire."""
        if self.response.is_done():
            return self.created_at + timedelta(minutes=15)
        else:
            return self.created_at + timedelta(seconds=3)

    def is_expired(self) -> bool:
        """:class:`bool` A boolean whether the interaction token is invalid or not."""
        return utils.utcnow() > self.expires_at

    def _set_application_command(
        self, app_cmd: Union[SlashApplicationSubcommand, BaseApplicationCommand]
    ):
        self.application_command = app_cmd

    @utils.cached_slot_property("_cs_channel")
    def channel(self) -> Optional[InteractionChannel]:
        """Optional[Union[:class:`abc.GuildChannel`, :class:`PartialMessageable`, :class:`Thread`]]: The channel the interaction was sent from.

        Note that due to a Discord limitation, DM channels are not resolved since there is
        no data to complete them. These are :class:`PartialMessageable` instead.
        """
        guild = self.guild
        channel = guild and guild._resolve_channel(self.channel_id)
        if channel is None:
            if self.channel_id is not None:
                type = ChannelType.text if self.guild_id is not None else ChannelType.private
                return PartialMessageable(state=self._state, id=self.channel_id, type=type)
            return None
        return channel

    @property
    def permissions(self) -> Permissions:
        """:class:`Permissions`: The resolved permissions of the member in the channel, including overwrites.

        In a non-guild context where this doesn't apply, an empty permissions object is returned.
        """
        return Permissions(self._permissions)

    @property
    def app_permissions(self) -> Permissions:
        """:class:`Permissions`: The resolved permissions of the bot in the channel, including overwrites.

        In a non-guild context where this doesn't apply, an empty permissions object is returned.
        """
        return Permissions(self._app_permissions)

    @utils.cached_slot_property("_cs_response")
    def response(self) -> InteractionResponse:
        """:class:`InteractionResponse`: Returns an object responsible for handling responding to the interaction.

        A response can only be done once. If secondary messages need to be sent, consider using :attr:`followup`
        instead.
        """
        return InteractionResponse(self)

    @utils.cached_slot_property("_cs_followup")
    def followup(self) -> Webhook:
        """:class:`Webhook`: Returns the follow up webhook for follow up interactions."""
        payload = {
            "id": self.application_id,
            "type": 3,
            "token": self.token,
        }
        return Webhook.from_state(data=payload, state=self._state)

    async def original_message(self) -> InteractionMessage:
        """|coro|

        Fetches the original interaction response message associated with the interaction.

        If the interaction response was :meth:`InteractionResponse.send_message` then this would
        return the message that was sent using that response. Otherwise, this would return
        the message that triggered the interaction.

        Repeated calls to this will return a cached value.

        Raises
        ------
        HTTPException
            Fetching the original response message failed.
        ClientException
            The channel for the message could not be resolved.

        Returns
        -------
        InteractionMessage
            The original interaction response message.
        """

        if self._original_message is not None:
            return self._original_message

        # TODO: fix later to not raise?
        channel = self.channel
        if channel is None:
            raise ClientException("Channel for message could not be resolved")

        adapter = async_context.get()
        data = await adapter.get_original_interaction_response(
            application_id=self.application_id,
            token=self.token,
            session=self._session,
        )
        state = _InteractionMessageState(self, self._state)
        message = InteractionMessage(state=state, channel=channel, data=data)  # type: ignore
        self._original_message = message
        return message

    async def edit_original_message(
        self,
        *,
        content: Optional[str] = MISSING,
        embeds: List[Embed] = MISSING,
        embed: Optional[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        attachments: List[Attachment] = MISSING,
        view: Optional[View] = MISSING,
        allowed_mentions: Optional[AllowedMentions] = None,
    ) -> InteractionMessage:
        """|coro|

        Edits the original interaction response message.

        This is a lower level interface to :meth:`InteractionMessage.edit` in case
        you do not want to fetch the message and save an HTTP request.

        This method is also the only way to edit the original message if
        the message sent was ephemeral.

        Parameters
        ----------
        content: Optional[:class:`str`]
            The content to edit the message with or ``None`` to clear it.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        file: :class:`File`
            The file to upload. This cannot be mixed with ``files`` parameter.
        files: List[:class:`File`]
            A list of files to send with the content. This cannot be mixed with the
            ``file`` parameter.
        attachments: List[:class:`Attachment`]
            A list of attachments to keep in the message. To keep existing attachments,
            you must fetch the message with :meth:`original_message` and pass
            ``message.attachments`` to this parameter.
        allowed_mentions: :class:`AllowedMentions`
            Controls the mentions being processed in this message.
            See :meth:`.abc.Messageable.send` for more information.
        view: Optional[:class:`~nextcord.ui.View`]
            The updated view to update this message with. If ``None`` is passed then
            the view is removed.

        Raises
        ------
        HTTPException
            Editing the message failed.
        Forbidden
            Edited a message that is not yours.
        InvalidArgument
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``.
        ValueError
            The length of ``embeds`` was invalid.

        Returns
        -------
        :class:`InteractionMessage`
            The newly edited message.
        """

        previous_mentions: Optional[AllowedMentions] = self._state.allowed_mentions
        params = handle_message_parameters(
            content=content,
            file=file,
            files=files,
            attachments=attachments,
            embed=embed,
            embeds=embeds,
            view=view,
            allowed_mentions=allowed_mentions,
            previous_allowed_mentions=previous_mentions,
        )
        adapter = async_context.get()
        data = await adapter.edit_original_interaction_response(
            self.application_id,
            self.token,
            session=self._session,
            payload=params.payload,
            multipart=params.multipart,
            files=params.files,
        )

        # The message channel types should always match
        message = InteractionMessage(state=self._state, channel=self.channel, data=data)  # type: ignore
        if view and not view.is_finished():
            self._state.store_view(view, message.id)
        return message

    async def delete_original_message(self, *, delay: Optional[float] = None) -> None:
        """|coro|

        Deletes the original interaction response message.

        This is a lower level interface to :meth:`InteractionMessage.delete` in case
        you do not want to fetch the message and save an HTTP request.

        Parameters
        ----------
        delay: Optional[:class:`float`]
            If provided, the number of seconds to wait before deleting the message.
            The waiting is done in the background and deletion failures are ignored.

        Raises
        ------
        HTTPException
            Deleting the message failed.
        Forbidden
            Deleted a message that is not yours.
        """
        adapter = async_context.get()
        delete_func = adapter.delete_original_interaction_response(
            self.application_id,
            self.token,
            session=self._session,
        )

        if delay is not None:

            async def inner_call(delay: float = delay):
                await asyncio.sleep(delay)
                try:
                    await delete_func
                except HTTPException:
                    pass

            asyncio.create_task(inner_call())
        else:
            await delete_func

    async def send(
        self,
        content: Optional[str] = None,
        *,
        embed: Embed = MISSING,
        embeds: List[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        view: View = MISSING,
        tts: bool = False,
        ephemeral: bool = False,
        delete_after: Optional[float] = None,
        allowed_mentions: AllowedMentions = MISSING,
    ) -> Union[PartialInteractionMessage, WebhookMessage]:
        """|coro|

        This is a shorthand function for helping in sending messages in
        response to an interaction. If the interaction has not been responded to,
        :meth:`InteractionResponse.send_message` is used. If the response
        :meth:`~InteractionResponse.is_done` then the message is sent
        via :attr:`Interaction.followup` using :class:`Webhook.send` instead.

        Raises
        ------
        HTTPException
            Sending the message failed.
        NotFound
            The interaction has expired or the interaction has been responded to
            but the followup webhook is expired.
        Forbidden
            The authorization token for the webhook is incorrect.
        InvalidArgument
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``.
        ValueError
            The length of ``embeds`` was invalid.

        Returns
        -------
        Union[:class:`PartialInteractionMessage`, :class:`WebhookMessage`]
            If the interaction has not been responded to, returns a :class:`PartialInteractionMessage`
            supporting only the :meth:`~PartialInteractionMessage.edit` and :meth:`~PartialInteractionMessage.delete`
            operations. To fetch the :class:`InteractionMessage` you may use :meth:`~PartialInteractionMessage.fetch`
            or :meth:`Interaction.original_message`.
            If the interaction has been responded to, returns the :class:`WebhookMessage`.
        """

        if not self.response.is_done():
            return await self.response.send_message(
                content=content,
                embed=embed,
                embeds=embeds,
                file=file,
                files=files,
                view=view,
                tts=tts,
                ephemeral=ephemeral,
                delete_after=delete_after,
                allowed_mentions=allowed_mentions,
            )
        return await self.followup.send(
            content=content,  # type: ignore
            embed=embed,
            embeds=embeds,
            file=file,
            files=files,
            view=view,
            tts=tts,
            ephemeral=ephemeral,
            delete_after=delete_after,
            allowed_mentions=allowed_mentions,
        )

    async def edit(self, *args, **kwargs) -> Optional[Message]:
        """|coro|

        This is a shorthand function for helping in editing messages in
        response to a component or modal submit interaction. If the
        interaction has not been responded to, :meth:`InteractionResponse.edit_message`
        is used. If the response :meth:`~InteractionResponse.is_done` then
        the message is edited via the :attr:`Interaction.message` using
        :meth:`Message.edit` instead.

        Raises
        ------
        HTTPException
            Editing the message failed.
        InvalidArgument
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``.
        TypeError
            An object not of type :class:`File` was passed to ``file`` or ``files``.
        HTTPException
            Editing the message failed.
        InvalidArgument
            :attr:`Interaction.message` was ``None``, this may occur in a :class:`Thread`
            or when the interaction is not a component or modal submit interaction.

        Returns
        -------
        Optional[:class:`Message`]
            The edited message. If the interaction has not yet been responded to,
            :meth:`InteractionResponse.edit_message` is used which returns
            a :class:`Message` or ``None`` corresponding to :attr:`Interaction.message`.
            Otherwise, the :class:`Message` is returned via :meth:`Message.edit`.
        """
        if not self.response.is_done():
            return await self.response.edit_message(*args, **kwargs)
        if self.message is not None:
            return await self.message.edit(*args, **kwargs)
        raise InvalidArgument(
            "Interaction.message is None, this method can only be used in "
            "response to a component or modal submit interaction."
        )


class InteractionResponse:
    """Represents a Discord interaction response.

    This type can be accessed through :attr:`Interaction.response`.

    .. versionadded:: 2.0
    """

    __slots__: Tuple[str, ...] = (
        "_responded",
        "_parent",
    )

    def __init__(self, parent: Interaction):
        self._parent: Interaction = parent
        self._responded: bool = False

    def is_done(self) -> bool:
        """:class:`bool`: Indicates whether an interaction response has been done before.

        An interaction can only be responded to once.
        """
        return self._responded

    async def defer(self, *, ephemeral: bool = False, with_message: bool = False) -> None:
        """|coro|

        Defers the interaction response.

        This is typically used when the interaction is acknowledged
        and a secondary action will be done later.

        Parameters
        ----------
        ephemeral: :class:`bool`
            Indicates whether the deferred message will eventually be ephemeral.
            This only applies for interactions of type :attr:`InteractionType.application_command` or when ``with_message`` is True
        with_message: :class:`bool`
            Indicates whether the response will be a message with thinking state (bot is thinking...).
            This is always True for interactions of type :attr:`InteractionType.application_command`.
            For interactions of type :attr:`InteractionType.component` this defaults to False.

            .. versionadded:: 2.0

        Raises
        ------
        HTTPException
            Deferring the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        defer_type: int = 0
        data: Optional[Dict[str, Any]] = None
        parent = self._parent
        if parent.type is InteractionType.application_command or with_message:
            defer_type = InteractionResponseType.deferred_channel_message.value
            if ephemeral:
                data = {"flags": 64}
        elif (
            parent.type is InteractionType.component or parent.type is InteractionType.modal_submit
        ):
            defer_type = InteractionResponseType.deferred_message_update.value

        if defer_type:
            adapter = async_context.get()
            await adapter.create_interaction_response(
                parent.id, parent.token, session=parent._session, type=defer_type, data=data
            )
            self._responded = True

    async def pong(self) -> None:
        """|coro|

        Pongs the ping interaction.

        This should rarely be used.

        Raises
        ------
        HTTPException
            Ponging the interaction failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        if parent.type is InteractionType.ping:
            adapter = async_context.get()
            await adapter.create_interaction_response(
                parent.id,
                parent.token,
                session=parent._session,
                type=InteractionResponseType.pong.value,
            )
            self._responded = True

    async def send_autocomplete(self, choices: Union[dict, list]) -> None:
        """|coro|

        Responds to this interaction by sending an autocomplete payload.

        Parameters
        ----------
        choices: Union[:class:`dict`, :class:`list`]
            The choices to send the user.
            If a :class:`dict` is given, each key-value pair is turned into a name-value pair. Name is what Discord
            shows the user, value is what Discord sends to the bot.
            If something not a :class:`dict`, such as a :class:`list`, is given, each value is turned into a duplicate
            name-value pair, where the display name and the value Discord sends back are the same.

        Raises
        ------
        HTTPException
            Sending the message failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)
        if not isinstance(choices, dict):
            choice_list = [{"name": choice, "value": choice} for choice in choices]
        else:
            choice_list = [{"name": key, "value": value} for key, value in choices.items()]

        payload = {"choices": choice_list}

        adapter = async_context.get()
        await adapter.create_interaction_response(
            self._parent.id,
            self._parent.token,
            session=self._parent._session,
            type=InteractionResponseType.application_command_autocomplete_result.value,
            data=payload,
        )
        self._responded = True

    async def send_message(
        self,
        content: Optional[Any] = None,
        *,
        embed: Embed = MISSING,
        embeds: List[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        view: View = MISSING,
        tts: bool = False,
        ephemeral: bool = False,
        delete_after: Optional[float] = None,
        allowed_mentions: Optional[AllowedMentions] = MISSING,
    ) -> PartialInteractionMessage:
        """|coro|

        Responds to this interaction by sending a message.

        Parameters
        ----------
        content: Optional[:class:`str`]
            The content of the message to send.
        embeds: List[:class:`Embed`]
            A list of embeds to send with the content. Maximum of 10. This cannot
            be mixed with the ``embed`` parameter.
        embed: :class:`Embed`
            The rich embed for the content to send. This cannot be mixed with
            ``embeds`` parameter.
        file: :class:`File`
            The file to upload.
        files: List[:class:`File`]
            A list of files to upload. Maximum of 10. This cannot be mixed with
            the ``file`` parameter.
        tts: :class:`bool`
            Indicates if the message should be sent using text-to-speech.
        view: :class:`nextcord.ui.View`
            The view to send with the message.
        ephemeral: :class:`bool`
            Indicates if the message should only be visible to the user who started the interaction.
            If a view is sent with an ephemeral message and it has no timeout set then the timeout
            is set to 15 minutes.
        delete_after: Optional[:class:`float`]
            If provided, the number of seconds to wait in the background
            before deleting the message we just sent. If the deletion fails,
            then it is silently ignored.
        allowed_mentions: :class:`AllowedMentions`
            Controls the mentions being processed in this message.
            See :meth:`.abc.Messageable.send` for more information.

        Raises
        ------
        HTTPException
            Sending the message failed.
        NotFound
            The interaction has expired. :meth:`InteractionResponse.defer` and
            :attr:`Interaction.followup` should be used if the interaction will take
            a while to respond.
        InvalidArgument
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``.
        TypeError
            An object not of type :class:`File` was passed to ``file`` or ``files``.
        ValueError
            The length of ``embeds`` was invalid.
        InteractionResponded
            This interaction has already been responded to before.

        Returns
        -------
        :class:`PartialInteractionMessage`
            An object supporting only the :meth:`~PartialInteractionMessage.edit` and :meth:`~PartialInteractionMessage.delete`
            operations. To fetch the :class:`InteractionMessage` you may use :meth:`PartialInteractionMessage.fetch`
            or :meth:`Interaction.original_message`.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        payload: Dict[str, Any] = {
            "tts": tts,
        }

        if embed is not MISSING and embeds is not MISSING:
            raise InvalidArgument("Cannot mix embed and embeds keyword arguments")

        if embed is not MISSING:
            embeds = [embed]

        if embeds:
            payload["embeds"] = [e.to_dict() for e in embeds]

        if file is not MISSING and files is not MISSING:
            raise InvalidArgument("Cannot mix file and files keyword arguments")

        if file is not MISSING:
            files = [file]

        if files and not all(isinstance(f, File) for f in files):
            raise TypeError("Files parameter must be a list of type File")

        if content is not None:
            payload["content"] = str(content)

        if ephemeral:
            payload["flags"] = 64

        if view is not MISSING:
            payload["components"] = view.to_components()

        if allowed_mentions is MISSING or allowed_mentions is None:
            if self._parent._state.allowed_mentions is not None:
                payload["allowed_mentions"] = self._parent._state.allowed_mentions.to_dict()
        else:
            if self._parent._state.allowed_mentions is not None:
                payload["allowed_mentions"] = self._parent._state.allowed_mentions.merge(
                    allowed_mentions
                ).to_dict()
            else:
                payload["allowed_mentions"] = allowed_mentions.to_dict()

        parent = self._parent
        adapter = async_context.get()
        try:
            await adapter.create_interaction_response(
                parent.id,
                parent.token,
                session=parent._session,
                type=InteractionResponseType.channel_message.value,
                data=payload,
                files=files,
            )
        finally:
            if files:
                for file in files:
                    file.close()

        if view is not MISSING:
            if ephemeral and view.timeout is None:
                view.timeout = 15 * 60.0

            self._parent._state.store_view(view)

        self._responded = True

        if delete_after is not None:
            await self._parent.delete_original_message(delay=delete_after)

        state = _InteractionMessageState(self._parent, self._parent._state)
        return PartialInteractionMessage(state)

    async def send_modal(self, modal: Modal) -> None:
        """|coro|

        Respond to this interaction by sending a modal.

        Parameters
        ----------
        modal: :class:`nextcord.ui.Modal`
            The modal to be sent in response to the interaction and which will
            be displayed on the user's screen.

        Raises
        ------
        HTTPException
            Sending the modal failed.
        InteractionResponded
            This interaction has already been responded to before.
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        adapter = async_context.get()
        await adapter.create_interaction_response(
            parent.id,
            parent.token,
            session=parent._session,
            type=InteractionResponseType.modal.value,
            data=modal.to_dict(),
        )

        self._responded = True

        self._parent._state.store_modal(modal, self._parent.user.id)  # type: ignore

    async def edit_message(
        self,
        *,
        content: Optional[Any] = MISSING,
        embed: Optional[Embed] = MISSING,
        embeds: List[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        attachments: List[Attachment] = MISSING,
        view: Optional[View] = MISSING,
        delete_after: Optional[float] = None,
    ) -> Optional[Message]:
        """|coro|

        Responds to this interaction by editing the message that the
        component or modal submit interaction originated from.

        Parameters
        ----------
        content: Optional[:class:`str`]
            The new content to replace the message with. ``None`` removes the content.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        file: :class:`File`
            The file to upload.
        files: List[:class:`File`]
            A list of files to upload. Maximum of 10. This cannot be mixed with
            the ``file`` parameter.
        attachments: List[:class:`Attachment`]
            A list of attachments to keep in the message. To keep all existing attachments,
            pass ``interaction.message.attachments``.
        view: Optional[:class:`~nextcord.ui.View`]
            The updated view to update this message with. If ``None`` is passed then
            the view is removed.
        delete_after: Optional[:class:`float`]
            If provided, the number of seconds to wait in the background
            before deleting the message we just sent. If the deletion fails,
            then it is silently ignored.


        Raises
        ------
        HTTPException
            Editing the message failed.
        InvalidArgument
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``.
        TypeError
            An object not of type :class:`File` was passed to ``file`` or ``files``.
        InteractionResponded
            This interaction has already been responded to before.

        Returns
        -------
        Optional[:class:`Message`]
            The message that was edited, or None if the :attr:`Interaction.message` is not found
            (this may happen if the interaction occurred in a :class:`Thread`).
        """
        if self._responded:
            raise InteractionResponded(self._parent)

        parent = self._parent
        msg = parent.message
        state = parent._state
        message_id = msg.id if msg else None

        payload = {}
        if content is not MISSING:
            if content is None:
                payload["content"] = None
            else:
                payload["content"] = str(content)

        if embed is not MISSING and embeds is not MISSING:
            raise InvalidArgument("Cannot mix both embed and embeds keyword arguments")

        if embed is not MISSING:
            if embed is None:
                embeds = []
            else:
                embeds = [embed]

        if embeds is not MISSING:
            payload["embeds"] = [e.to_dict() for e in embeds]

        if file is not MISSING and files is not MISSING:
            raise InvalidArgument("Cannot mix file and files keyword arguments")

        if file is not MISSING:
            files = [file]

        if files and not all(isinstance(f, File) for f in files):
            raise TypeError("Files parameter must be a list of type File")

        if attachments is not MISSING:
            payload["attachments"] = [a.to_dict() for a in attachments]

        if view is not MISSING:
            if message_id is not None:
                state.prevent_view_updates_for(message_id)
            if view is None:
                payload["components"] = []
            else:
                payload["components"] = view.to_components()

        adapter = async_context.get()
        try:
            await adapter.create_interaction_response(
                parent.id,
                parent.token,
                session=parent._session,
                type=InteractionResponseType.message_update.value,
                data=payload,
                files=files,
            )
        finally:
            if files:
                for file in files:
                    file.close()

        if view and not view.is_finished() and message_id is not None:
            state.store_view(view, message_id)

        self._responded = True

        if delete_after is not None:
            await parent.delete_original_message(delay=delete_after)

        return state._get_message(message_id)


class _InteractionMessageState:
    __slots__ = ("_parent", "_interaction")

    def __init__(self, interaction: Interaction, parent: ConnectionState):
        self._interaction: Interaction = interaction
        self._parent: ConnectionState = parent

    def _get_guild(self, guild_id):
        return self._parent._get_guild(guild_id)

    def store_user(self, data):
        return self._parent.store_user(data)

    def create_user(self, data):
        return self._parent.create_user(data)

    @property
    def http(self):
        return self._parent.http

    def __getattr__(self, attr):
        return getattr(self._parent, attr)


class _InteractionMessageMixin:
    __slots__ = ()
    _state: _InteractionMessageState

    async def edit(
        self,
        content: Optional[str] = MISSING,
        embeds: List[Embed] = MISSING,
        embed: Optional[Embed] = MISSING,
        file: File = MISSING,
        files: List[File] = MISSING,
        attachments: List[Attachment] = MISSING,
        view: Optional[View] = MISSING,
        allowed_mentions: Optional[AllowedMentions] = None,
        delete_after: Optional[float] = None,
    ) -> InteractionMessage:
        """|coro|

        Edits the message.

        Parameters
        ----------
        content: Optional[:class:`str`]
            The content to edit the message with or ``None`` to clear it.
        embeds: List[:class:`Embed`]
            A list of embeds to edit the message with.
        embed: Optional[:class:`Embed`]
            The embed to edit the message with. ``None`` suppresses the embeds.
            This should not be mixed with the ``embeds`` parameter.
        file: :class:`File`
            The file to upload. This cannot be mixed with ``files`` parameter.
        files: List[:class:`File`]
            A list of files to send with the content. This cannot be mixed with the
            ``file`` parameter.
        attachments: List[:class:`Attachment`]
            A list of attachments to keep in the message. To keep existing attachments,
            you must fetch the message with :meth:`original_message` and pass
            ``message.attachments`` to this parameter.
        allowed_mentions: :class:`AllowedMentions`
            Controls the mentions being processed in this message.
            See :meth:`.abc.Messageable.send` for more information.
        view: Optional[:class:`~nextcord.ui.View`]
            The updated view to update this message with. If ``None`` is passed then
            the view is removed.
        delete_after: Optional[:class:`float`]
            If provided, the number of seconds to wait in the background
            before deleting the message we just edited. If the deletion fails,
            then it is silently ignored.

        Raises
        ------
        HTTPException
            Editing the message failed.
        Forbidden
            Edited a message that is not yours.
        InvalidArgument
            You specified both ``embed`` and ``embeds`` or ``file`` and ``files``.
        ValueError
            The length of ``embeds`` was invalid.

        Returns
        -------
        :class:`InteractionMessage`
            The newly edited message.
        """
        message = await self._state._interaction.edit_original_message(
            content=content,
            embeds=embeds,
            embed=embed,
            file=file,
            files=files,
            attachments=attachments,
            view=view,
            allowed_mentions=allowed_mentions,
        )

        if delete_after is not None:
            await self.delete(delay=delete_after)

        return message

    async def delete(self, *, delay: Optional[float] = None) -> None:
        """|coro|

        Deletes the message.

        Parameters
        ----------
        delay: Optional[:class:`float`]
            If provided, the number of seconds to wait before deleting the message.
            The waiting is done in the background and deletion failures are ignored.

        Raises
        ------
        Forbidden
            You do not have proper permissions to delete the message.
        NotFound
            The message was deleted already.
        HTTPException
            Deleting the message failed.
        """

        await self._state._interaction.delete_original_message(delay=delay)


class PartialInteractionMessage(_InteractionMessageMixin):
    """Represents the original interaction response message when only the
    application state and interaction token are available.

    This allows you to edit or delete the message associated with
    the interaction response. This object is returned when responding to
    an interaction with :meth:`InteractionResponse.send_message`.

    This does not support most attributes and methods of :class:`nextcord.Message`.
    The :meth:`~PartialInteractionMessage.fetch` method can be used to
    retrieve the full :class:`InteractionMessage` object.

    .. container:: operations

        .. describe:: x == y

            Checks if two partial interaction messages are equal.

        .. describe:: x != y

            Checks if two partial interaction messages are not equal.

        .. describe:: hash(x)

            Returns the partial interaction message's hash.

    .. versionadded:: 2.0

    .. versionchanged:: 2.1
        :class:`PartialInteractionMessage` is now hashable by :attr:`Interaction.id`.

    .. note::
        The hash of a :class:`PartialInteractionMessage` is the same as the hash of the
        :class:`Interaction` that it is associated with but not that of the full :class:`InteractionMessage`.
    """

    def __init__(self, state: _InteractionMessageState):
        self._state = state

    async def fetch(self) -> InteractionMessage:
        """|coro|

        Fetches the original interaction response message associated with the interaction.

        Repeated calls to this will return a cached value.

        Raises
        ------
        HTTPException
            Fetching the original response message failed.
        ClientException
            The channel for the message could not be resolved.

        Returns
        -------
        InteractionMessage
            The original interaction response message.
        """
        return await self._state._interaction.original_message()

    @property
    def author(self) -> Optional[Union[Member, ClientUser]]:
        """Optional[Union[:class:`Member`, :class:`ClientUser`]]: The client that responded to the interaction.

        If the interaction was in a guild, this is a :class:`Member` representing the client.
        Otherwise, this is a :class:`ClientUser`.
        """
        return self.guild.me if self.guild else self._state._interaction.client.user

    @property
    def channel(self) -> Optional[InteractionChannel]:
        """Optional[Union[:class:`abc.GuildChannel`, :class:`PartialMessageable`, :class:`Thread`]]: The channel the interaction was sent from.

        Note that due to a Discord limitation, DM channels are not resolved since there is
        no data to complete them. These are :class:`PartialMessageable` instead.
        """
        return self._state._interaction.channel

    @property
    def guild(self) -> Optional[Guild]:
        """Optional[:class:`Guild`]: The guild the interaction was sent from."""
        return self._state._interaction.guild

    def __repr__(self):
        return f"<{self.__class__.__name__} author={self.author!r} channel={self.channel!r} guild={self.guild!r}>"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, PartialInteractionMessage)
            and self._state._interaction == other._state._interaction
        )

    def __ne__(self, other: object) -> bool:
        if isinstance(other, PartialInteractionMessage):
            return self._state._interaction != other._state._interaction
        return True

    def __hash__(self) -> int:
        return hash(self._state._interaction)


class InteractionMessage(_InteractionMessageMixin, Message):
    """Represents the original interaction response message.

    To retrieve this object see :meth:`PartialInteractionMessage.fetch`
    or :meth:`Interaction.original_message`.

    This inherits from :class:`nextcord.Message` with changes to
    :meth:`edit` and :meth:`delete` to work with the interaction response.

    .. versionadded:: 2.0
    """

    pass



"""
The MIT License (MIT)

Copyright (c) 2022-present tag-epic

Permission is hereby granted, free of charge, to any person obtaining a
copy of this software and associated documentation files (the "Software"),
to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense,
and/or sell copies of the Software, and to permit persons to whom the
Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, List, Optional

from .enums import (
    AutoModerationActionType,
    AutoModerationEventType,
    AutoModerationTriggerType,
    KeywordPresetType,
    try_enum,
)
from .errors import InvalidArgument
from .mixins import Hashable
from .object import Object
from .utils import MISSING, get_as_snowflake

if TYPE_CHECKING:
    from .abc import GuildChannel, Snowflake
    from .guild import Guild
    from .member import Member
    from .message import Message
    from .role import Role
    from .state import ConnectionState
    from .types.auto_moderation import (
        AutoModerationAction as AutoModerationActionPayload,
        AutoModerationActionExecution as ActionExecutionPayload,
        AutoModerationActionMetadata as ActionMetadataPayload,
        AutoModerationRule as AutoModerationRulePayload,
        AutoModerationRuleModify,
        AutoModerationTriggerMetadata as TriggerMetadataPayload,
    )

__all__ = (
    "AutoModerationTriggerMetadata",
    "AutoModerationActionMetadata",
    "AutoModerationAction",
    "AutoModerationRule",
    "AutoModerationActionExecution",
)


class AutoModerationTriggerMetadata:
    """Represents data about an auto moderation trigger rule.

    .. versionadded:: 2.1

    Attributes
    ----------
    keyword_filter: List[:class:`str`]
        A list of substrings which will be searched for in content.

        .. note::

            This is ``None`` and cannot be provided if the trigger type of the rule is not
            :attr:`AutoModerationTriggerType.keyword`.
    presets: List[:class:`KeywordPresetType`]
        A list of Discord pre-defined wordsets which will be searched for in content.

        .. note::

            This is ``None`` and cannot be provided if the trigger type of the rule is not
            :attr:`AutoModerationTriggerType.keyword_preset`.
    allow_list: List[:class:`str`]
        A list of exempt strings that will not trigger the preset type.

        .. note::

            This is ``None`` and cannot be provided if the trigger type of the rule is not
            :attr:`AutoModerationTriggerType.keyword_preset`.

        .. warning::

            Wildcard syntax (`*`) is not supported here.
    mention_total_limit: Optional[:class:`int`]
        The maximum amount of mentions allowed in a messsage.

        .. versionadded:: 2.3

        .. note::

            This is ``None`` and cannot be provided if the trigger type of this rule is not :attr:`AutoModerationTriggerType.mention_spam`.
    """

    __slots__ = ("keyword_filter", "presets", "allow_list", "mention_total_limit")

    def __init__(
        self,
        *,
        keyword_filter: Optional[List[str]] = None,
        presets: Optional[List[KeywordPresetType]] = None,
        allow_list: Optional[List[str]] = None,
        mention_total_limit: Optional[int] = None,
    ) -> None:
        self.keyword_filter: Optional[List[str]] = keyword_filter
        self.presets: Optional[List[KeywordPresetType]] = presets
        self.allow_list: Optional[List[str]] = allow_list
        self.mention_total_limit: Optional[int] = mention_total_limit

    @classmethod
    def from_data(cls, data: TriggerMetadataPayload):
        keyword_filter = data.get("keyword_filter")
        presets = (
            [try_enum(KeywordPresetType, preset) for preset in data["presets"]]
            if "presets" in data
            else None
        )
        allow_list = data.get("allow_list")
        mention_total_limit = data.get("mention_total_limit")

        return cls(
            keyword_filter=keyword_filter,
            presets=presets,
            allow_list=allow_list,
            mention_total_limit=mention_total_limit,
        )

    @property
    def payload(self) -> TriggerMetadataPayload:
        payload: TriggerMetadataPayload = {}

        if self.keyword_filter is not None:
            payload["keyword_filter"] = self.keyword_filter

        if self.presets is not None:
            payload["presets"] = [enum.value for enum in self.presets]

        if self.allow_list is not None:
            payload["allow_list"] = self.allow_list

        if self.mention_total_limit is not None:
            payload["mention_total_limit"] = self.mention_total_limit

        return payload


class AutoModerationActionMetadata:
    """Represents additional data that is used when an action is executed.

    .. versionadded:: 2.1

    Attributes
    ----------
    channel: Optional[:class:`abc.Snowflake`]
        The channel to which message content should be logged.

        .. note::

            This is ``None`` and cannot be provided if the action type of the rule is not
            :attr:`AutoModerationActionType.send_alert_message`.
    duration_seconds: Optional[:class:`int`]
        The duration of the timeout in seconds.

        .. note::

            This is ``None`` and cannot be provided if the action type of the rule is not
            :attr:`AutoModerationActionType.timeout`.

        .. note::

            The maximum value that can be used is ``2419200`` seconds (4 weeks)
    """

    __slots__ = ("channel_id", "duration_seconds")

    def __init__(
        self, *, channel: Optional[Snowflake] = None, duration_seconds: Optional[int] = None
    ):
        self.channel_id: Optional[int] = channel.id if channel is not None else None
        self.duration_seconds: Optional[int] = duration_seconds

    @classmethod
    def from_data(cls, data: ActionMetadataPayload):
        channel_id = get_as_snowflake(data, "channel_id")
        channel = Object(id=channel_id) if channel_id is not None else None
        duration_seconds = data.get("duration_seconds")

        return cls(channel=channel, duration_seconds=duration_seconds)

    @property
    def payload(self) -> ActionMetadataPayload:
        payload: ActionMetadataPayload = {}

        if self.channel_id is not None:
            payload["channel_id"] = self.channel_id

        if self.duration_seconds is not None:
            payload["duration_seconds"] = self.duration_seconds

        return payload


class AutoModerationAction:
    """Represents an auto moderation action that will execute whenever a rule is triggered.

    .. versionadded:: 2.1

    Parameters
    ----------
    type: :class:`AutoModerationActionType`
        The type to use for this action.
    metadata: :class:`AutoModerationActionMetadata`
        The additional data to use during execution of this action.

    Attributes
    ----------
    type: :class:`AutoModerationActionType`
        The type of this action.
    metadata: :class:`AutoModerationActionMetadata`
        The additional metadata needed during execution for this specific action type.
    """

    __slots__ = ("type", "metadata")

    def __init__(
        self,
        *,
        type: AutoModerationActionType,
        metadata: Optional[AutoModerationActionMetadata] = None,
    ) -> None:
        self.type: AutoModerationActionType = type
        self.metadata: Optional[AutoModerationActionMetadata] = metadata

    @classmethod
    def from_data(cls, data: AutoModerationActionPayload) -> AutoModerationAction:
        type = try_enum(AutoModerationActionType, data["type"])
        metadata = AutoModerationActionMetadata.from_data(data.get("metadata", {}))
        return cls(type=type, metadata=metadata)

    @property
    def payload(self) -> AutoModerationActionPayload:
        data: AutoModerationActionPayload = {
            "type": self.type.value,
        }

        if self.metadata is not None:
            data["metadata"] = self.metadata.payload

        return data


class AutoModerationRule(Hashable):
    """Represents a Discord auto moderation rule.

    .. versionadded:: 2.1

    .. container:: operations

        .. describe:: x == y

            Checks if two rules are equal.

        .. describe:: x != y

            Checks if two rules are not equal.

        .. describe:: hash(x)

            Returns the rules's hash.

        .. describe:: str(x)

            Returns the rules's name.

    Attributes
    ----------
    id: :class:`int`
        The rule's unique ID.
    guild_id: :class:`int`
        The guild's unique ID which this rule belongs to.
    guild: Optional[:class:`Guild`]
        The guild which this rule belongs to, if found in cache.
    name: :class:`str`
        The rule's name.
    creator_id: :class:`int`
        The user's unique ID which first created this rule.
    creator: Optional[:class:`Member`]
        The member which first created this rule, if found in cache.
    event_type: :class:`AutoModerationEventType`
        The event context in which this rule is checked.
    trigger_type: :class:`AutoModerationTriggerType`
        The type of content that can trigger this rule.
    trigger_metadata: :class:`AutoModerationTriggerMetadata`
        Additional data that is used when checking if this rule is triggered.
    actions: List[:class:`AutoModerationAction`]
        The actions which will execute when the rule is triggered.
    enabled: :class:`bool`
        Whether the rule is enabled.
    exempt_role_ids: List[:class:`int`]
        The role ids that should not be affected by the rule.
    exempt_channel_ids: List[:class:`int`]
        The channel ids that should not be affected by the rule.
    exempt_roles: List[:class:`Role`]
        The roles that should not be affected by the rule, if found in cache.
    exempt_channels: List[:class:`abc.GuildChannel`]
        The channels that should not be affected by the rule, if found in cache.
    """

    __slots__ = (
        "_state",
        "id",
        "guild_id",
        "guild",
        "name",
        "creator_id",
        "creator",
        "event_type",
        "trigger_type",
        "trigger_metadata",
        "actions",
        "enabled",
        "exempt_role_ids",
        "exempt_channel_ids",
        "exempt_roles",
        "exempt_channels",
    )

    def __init__(self, *, data: AutoModerationRulePayload, state: ConnectionState):
        self._state = state
        self.id: int = int(data["id"])
        self.guild_id: int = int(data["guild_id"])
        self.guild: Optional[Guild] = state._get_guild(self.guild_id)
        self.name: str = data["name"]
        self.creator_id: int = int(data["creator_id"])
        self.creator: Optional[Member] = (
            self.guild.get_member(self.creator_id) if self.guild is not None else None
        )
        self.event_type: AutoModerationEventType = try_enum(
            AutoModerationEventType, data["event_type"]
        )
        self.trigger_type: AutoModerationTriggerType = try_enum(
            AutoModerationTriggerType, data["trigger_type"]
        )
        self.trigger_metadata: AutoModerationTriggerMetadata = (
            AutoModerationTriggerMetadata.from_data(data["trigger_metadata"])
        )
        self.actions: List[AutoModerationAction] = [
            AutoModerationAction.from_data(data=action) for action in data["actions"]
        ]
        self.enabled: bool = data["enabled"]
        self.exempt_role_ids: List[int] = [int(role_id) for role_id in data["exempt_roles"]]
        self.exempt_channel_ids: List[int] = [
            int(channel_id) for channel_id in data["exempt_channels"]
        ]
        self.exempt_roles: List[Optional[Role]] = (
            [self.guild.get_role(role_id) for role_id in self.exempt_role_ids]
            if self.guild is not None
            else []
        )
        self.exempt_channels: List[Optional[GuildChannel]] = (
            [self.guild.get_channel(channel_id) for channel_id in self.exempt_channel_ids]
            if self.guild is not None
            else []
        )

    def __str__(self) -> str:
        return self.name

    async def delete(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Delete the current rule from its guild.

        Requires the :attr:`~Permissions.manage_guild` permission.

        Parameters
        ----------
        reason: Optional[:class:`str`]
            The reason for deleting this rule. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You do not have permission to delete this rule.
        HTTPException
            Deleting the rule failed.
        """

        await self._state.http.delete_auto_moderation_rule(self.guild_id, self.id, reason=reason)

    async def edit(
        self,
        *,
        name: str = MISSING,
        event_type: AutoModerationEventType = MISSING,
        trigger_metadata: AutoModerationTriggerMetadata = MISSING,
        actions: List[AutoModerationAction] = MISSING,
        enabled: bool = MISSING,
        exempt_roles: List[Snowflake] = MISSING,
        exempt_channels: List[Snowflake] = MISSING,
        reason: Optional[str] = None,
    ) -> AutoModerationRule:
        """Modify the current auto moderation rule.

        Requires the :attr:`~Permissions.manage_guild` permission.

        Parameters
        ----------
        name: :class:`str`
            The new name for this rule.
        event_type: :class:`AutoModerationEventType`
            The new event context in which this rule is checked.
        trigger_metadata: :class:`AutoModerationTriggerMetadata`
            The new additional data that is used when checking if this rule is triggered.
        actions: List[:class:`AutoModerationAction`]
            The new actions which will execute when the rule is triggered.
        enabled: :class:`bool`
            Whether the rule is enabled.
        exempt_roles: List[:class:`abc.Snowflake`]
            The new roles that should not be affected by the rule.
        exempt_channels: List[:class:`abc.Snowflake`]
            The new channels that should not be affected by the rule.
        reason: Optional[:class:`str`]
            The reason for editing this rule. Shows in the audit log.

        Raises
        ------
        Forbidden
            You do not have permission to edit this rule.
        HTTPException
            Editing the rule failed.
        InvalidArgument
            An incorrect type was passed.

        Returns
        -------
        :class:`AutoModerationRule`
            The newly edited auto moderation rule.
        """

        payload: AutoModerationRuleModify = {}

        if name is not MISSING:
            payload["name"] = name

        if event_type is not MISSING:
            if not isinstance(event_type, AutoModerationEventType):
                raise InvalidArgument("event_type must be of type AutoModerationEventType")

            payload["event_type"] = event_type.value

        if trigger_metadata is not MISSING:
            if not isinstance(trigger_metadata, AutoModerationTriggerMetadata):
                raise InvalidArgument(
                    "trigger_metadata must be of type AutoModerationTriggerMetadata"
                )

            payload["trigger_metadata"] = trigger_metadata.payload

        if actions is not MISSING:
            payload["actions"] = [action.payload for action in actions]

        if enabled is not MISSING:
            payload["enabled"] = enabled

        if exempt_roles is not MISSING:
            payload["exempt_roles"] = [str(role.id) for role in exempt_roles]

        if exempt_channels is not MISSING:
            payload["exempt_channels"] = [str(channel.id) for channel in exempt_channels]

        data = await self._state.http.modify_auto_moderation_rule(
            self.guild_id, self.id, data=payload, reason=reason
        )
        return AutoModerationRule(data=data, state=self._state)


class AutoModerationActionExecution:
    """Represents the execution of an auto moderation action

    .. versionadded:: 2.1

    Attributes
    ----------
    guild_id: :class:`int`
        The guild ID where this action was executed.
    guild: Optional[:class:`Guild`]
        The guild where this action was executed, if it was found in cache.
    channel_id: Optional[:class:`int`]
        The channel ID where this action was executed, if applicable.
    channel: Optional[:class:`abc.GuildChannel`]
        The channel where this action was executed, if applicable and found in cache.
    message_id: Optional[:class:`int`]
        The message ID that executed this action, if it was not blocked.
    message: Optional[:class:`Message`]
        The message that executed this action, if it was not blocked and found in cache.
    alert_system_message_id: Optional[:class:`int`]
        The ID of the system alert message, if sent.
    alert_system_message: Optional[:class:`Message`]
        The system alert message, if sent and found in cache.
    action: :class:`AutoModerationAction`
        The action that was executed.
    rule_id: :class:`int`
        The id of the rule that was executed.
    rule_trigger_type: :class:`AutoModerationTriggerType`
        The type of rule that was executed.
    member_id: :class:`int`
        The ID of the user that triggered this action.
    member: Optional[:class:`Member`]
        The member that triggered this action, if found in cache.
    content: :class:`str`
        The content the user sent in the message

        .. note::

            This requires :attr:`Intents.message_content` to not be empty.
    matched_keyword: Optional[:class:`str`]
        The keyword configured in the rule that matched this message, if applicable.
    matched_content: Optional[:class:`str`]
        The content in the message that matched the keyword, if applicable.

        .. note::

            This requires :attr:`Intents.message_conrent` to not be empty.
    """

    def __init__(self, *, data: ActionExecutionPayload, state: ConnectionState) -> None:
        self.guild_id: int = int(data["guild_id"])
        self.guild: Optional[Guild] = state._get_guild(self.guild_id)
        self.channel_id: Optional[int] = get_as_snowflake(data, "channel_id")
        if self.guild is not None and self.channel_id is not None:
            self.channel: Optional[GuildChannel] = self.guild.get_channel(self.channel_id)
        else:
            self.channel: Optional[GuildChannel] = None
        self.message_id: Optional[int] = get_as_snowflake(data, "message_id")
        self.message: Optional[Message] = state._get_message(self.message_id)
        self.alert_system_message_id: Optional[int] = get_as_snowflake(
            data, "alert_system_message_id"
        )
        self.alert_system_message: Optional[Message] = state._get_message(
            self.alert_system_message_id
        )
        self.action: AutoModerationAction = AutoModerationAction.from_data(data=data["action"])
        self.rule_id: int = int(data["rule_id"])
        self.rule_trigger_type: AutoModerationTriggerType = try_enum(
            AutoModerationTriggerType, data["rule_trigger_type"]
        )
        self.member_id: int = int(data["user_id"])
        self.member: Optional[Member] = (
            self.guild.get_member(self.member_id) if self.guild is not None else None
        )
        self.content: str = data.get("content", "")
        self.matched_keyword: Optional[str] = data["matched_keyword"]
        self.matched_content: Optional[str] = data.get("matched_content", "")
